/**
 * @file modelconverter.hpp
 * @date 2015-06-21
 * @brief Tv\[XR[h
 */
// for June2015v2 

#pragma once

#include "mqo.hpp"
#include "pmd.hpp"

/// <summary>ZW]VCX^XԂ</summary>
Vec3 rev_z(const Vec3& v) {
	Vec3 ret;
	ret.x = v.x;
	ret.y = v.y;
	ret.z = -v.z;
	return ret;
}

/// <summary>ItZbgƌBގƁB</summary>
struct VertexGroup {
	/// <summary>MeshData̖ʒ_CfbNX̃ItZbg</summary>
	int offset = 0;
	/// <summary>MeshData̖ʒ_CfbNX̌</summary>
	int num() const {
		return indices.size();
	};

	/// <summary>ʒ_</summary>
	Array<int> indices;
};

/// <summary>1[V</summary>
struct MotionNode {
	/// <summary>O</summary>
	String name;

	/// <summary>eCfbNX
	/// -1̏ꍇ̓fm[he</summary>
	int parentIndex = -1;

	/// <summary>Sivł̕sړ</summary>
	Vec3 translation = { 0, 0, 0 };
	/// <summary>Sivł̃[J]
	/// rotationvZ</summary>
	Quaternion rot;

	/// <summary>ʒuinO[o</summary>
	Vec3 globalMov = { 0, 0, 0 };
	/// <summary>]inO[o</summary>
	Quaternion globalRot = Quaternion::Identity();
	/// <summary>XP[inO[o(Ή)</summary>
	Vec3 globalRate = { 1, 1, 1 };

	/// <summary>[VƂĔf镪</summary>
	Vec3 motionMov = { 0, 0, 0 };
	/// <summary>[VƂĔf镪</summary>
	Quaternion motionRot = Quaternion::Identity();
	/// <summary>[VƂĔf镪(Ή)</summary>
	Vec3 motionScale = { 1, 1, 1 };

	/// <summary>ƃ[VfZoړ</summary>
	Vec3 curMov = { 0, 0, 0 };
	/// <summary>ƃ[VfZo]B
	/// ɉ]B</summary>
	Quaternion curRot = Quaternion::Identity();
	/// <summary>ƃ[VfZogk</summary>
	Vec3 curScale = { 1, 1, 1 };

	/// <summary>ގƓ</summary>
	Array<VertexGroup> groups;

	/// <summary>vZ̍ŏIs</summary>
	Mat4x4 GetCurMat() const {
		Quaternion revGlobalRot = Quaternion(-globalRot.component.fv.x,
			-globalRot.component.fv.y,
			-globalRot.component.fv.z,
			globalRot.component.fv.w);
		Mat4x4 ret = Mat4x4::Translate(-globalMov).rotated(revGlobalRot).rotated(curRot).translated(curMov);
		return ret;
	}
};

/// <summary>.mqo1ObjectɑΉ</summary>
struct ObjectNode : public MotionNode {
	/// <summary>.mqołdepth
	/// -1̓fS</summary>
	int depth = -1;
	/// <summary>XP[͔Ή</summary>
	Vec3 scale = {1,1,1};

	/// <summary>SivłRoll,Pitch,Yaw</summary>
	Vec3 rotation = {0,0,0};


	/// <summary>\
	/// 0̏ꍇ͎q</summary>
	int childIndex = 0;

	/// <summary>_Ɩʒ_CfbNX
	/// ̖ʒ_CfbNX͍ގ</summary>
	MeshData meshData;
};

/// <summary>1̃CX^X</summary>
struct MotionInstance {
	/// <summary>S</summary>
	MotionNode modelNode;
	/// <summary>IuWFNg</summary>
	Array<MotionNode> nodes;

	/// <summary>ǉCfbNXB
	/// 1ft@C̑ΉCfbNXi[B</summary>
	Optional<int> optIdx;

	/// <summary>Object̏ԂvZB
	/// scale͖ 1 1 1 Ƃ݂Ȃ</summary>
	void ComputeCurCoordinates() {
		int num = nodes.size();
		for (int i = 0; i < num; ++i) {
			MotionNode& node = nodes[i];
			// ύXȂ悤ɃRs[
			MotionNode n = this->modelNode;
			int parent = node.parentIndex;
			if (parent >= 0 && parent < num) {
				n = nodes[parent];
			}

			node.curMov = n.curMov
				+ n.curRot * (node.translation + node.motionMov);
			node.curRot = n.curRot * node.rot * node.motionRot;

			//Print(L"ʒu, ", node.curMov);
			//Println(L", ], ", node.curRot);
		}
	}
};

/// <summary>.mqoƃbVf[^Ԃ̕ϊ
/// Siv3D\sB</summary>
class ModelConverter {
public:
	/// <summary>ގ</summary>
	Array<Material> mats;
	/// <summary>eNX`</summary>
	Array<DynamicTexture> texs;

	/// <summary>bV</summary>
	Array<DynamicMesh> meshes;

	/// <summary>IuWFNgƓ</summary>
	Array<ObjectNode> nodes;

	/// <summary>ŏi</summary>
	ObjectNode modelNode;
	/// <summary>t@C̃pX</summary>
	String path;
	/// <summary>gpBWATCY̕ϊ</summary>
	//double rate = 1.0;

	/// <summary>zNA</summary>
	void clear() {
		this->mats.clear();
		this->texs.clear();
		this->meshes.clear();
		this->nodes.clear();
		this->modelNode = ObjectNode();
		this->path.clear();
	}

	/// <summary>bVf[^𐶐</summary>
	/// <param name="f">Kp</param>
	/// <param name="rate">Wl̔{w肷</param>
	/// <param name="option">ʍގ̈A</param>
	bool ApplyMqo(const MqoFile& f,
			double rate,
			const Array<String>& options) {
		bool ret = true;
		this->clear();

		bool addFaceForDouble = false; // ʍގ̏ꍇɖʒǉ
		for (const auto& option: options) {
			if (option == L"addFaceForDouble") {
				addFaceForDouble = true;
			}
		}

		this->path = f.path;

		// ގ̂߂̃eNX`
		String dir = s3d::FileSystem::ParentPath(f.path);

		int matNum = f.materials.size();
		for (const auto& m : f.materials) {
			Material mat;
			mat.diffuse = m.col * m.dif;
			mat.specular = m.col * m.spc;
			mat.shininess = m.power;
			mat.emission = m.col * m.emi;
			mat.ambient = m.col * m.amb;
			//mat.illum = 1;
			mat.diffuseTextureName = m.tex.value_or(L"");
			mat.ambientTextureName = L"";
			mat.normalTextureName = m.bump.value_or(L"");
			mat.name = m.name;
			this->mats.push_back(mat);

			bool flag = false;
			if (m.tex) {
				auto name = m.tex.value();
				auto fullPath = name;
				if (fullPath.indexOf(L":") != 1) {
					fullPath = dir + L"/" + name;
					if (FileSystem::Exists(fullPath) == false) {
						fullPath = dir + L"/Texture/" + name;
					}
				}

				Image img(fullPath);
				if (img) {
					DynamicTexture dtex(img);
					if (dtex) {
						this->texs.push_back(dtex);
						flag = true;
					}
				}
			}
			if (flag == false) {
				DynamicTexture dtex; // ftHg
				this->texs.push_back(dtex);
			}
		}

		for (int i = 0; i < f.objects.size(); ++i) {
			MqoObject obj = f.objects[i];
			ObjectNode node;
			node.name = obj.name;
			node.depth = obj.depth;
			node.translation = rev_z(obj.translation) * rate;
			node.rotation = { Radians(obj.rotation.z),
				-Radians(obj.rotation.y),
				-Radians(obj.rotation.x) };
			node.scale = obj.scale;

			int depth = obj.depth;
			int index = -1;
			// ԋ߂1T
			if (depth > 0) {
				for (int j = 0; j < i; ++j) {
					if (f.objects[j].depth == depth - 1) {
						index = j;
					}
				}
			}
			node.parentIndex = index;
			node.childIndex = 0;
			if (index >= 0 && index < this->nodes.size()) {
				if (this->nodes[index].childIndex <= 0) {
					this->nodes[index].childIndex = i;
				}
			}

			node.rot = Quaternion::RollPitchYaw(
				node.rotation.x,
				node.rotation.y,
				node.rotation.z);

			// ގ
			for (int j = 0; j < matNum; ++j) {
				VertexGroup gr;
				gr.offset = 0;
				gr.indices.clear();
				node.groups.push_back(gr);
			}

			Vec2 duv(-0.5, -0.5);
			// 򂷂
			if (obj.shading == 0) { // oŎOp`S@

				for (const auto& face : obj.faces) {
					int matIndex = face.materialIndex;
					if (matIndex < 0 || matIndex >= matNum) {
						continue;
					}
					VertexGroup& gr = node.groups[matIndex];

					bool dbls = f.materials[matIndex].dbls;
					bool addDoubleFace = false;
					if (dbls && addFaceForDouble) {
						addDoubleFace = true;
					}

					int n = face.indices.size();
					int uvNum = face.uvs.size();
					int v0, v1, v2;
					if (n >= 3) {
						v0 = face.indices[0];
						for (int i = 0; i < n - 2; ++i) {
							int baseIndex = node.meshData.vertices.size();
							v1 = face.indices[i + 1];
							v2 = face.indices[i + 2];
							gr.indices.push_back(baseIndex + 0);
							gr.indices.push_back(baseIndex + 1);
							gr.indices.push_back(baseIndex + 2);

							if (addDoubleFace) {
								gr.indices.push_back(baseIndex + 2);
								gr.indices.push_back(baseIndex + 1);
								gr.indices.push_back(baseIndex + 0);
							}

							MeshVertex mv0;
							MeshVertex mv1;
							MeshVertex mv2;
							mv0.position = rev_z(obj.vertices[v0]) * rate;
							if (0 < uvNum) {
								mv0.texcoord = face.uvs[0];
							}
							else {
								mv0.texcoord = duv;
							}
							mv1.position = rev_z(obj.vertices[v1]) * rate;
							if (i + 1 < uvNum) {
								mv1.texcoord = face.uvs[i + 1];
							}
							else {
								mv1.texcoord = duv;
							}
							mv2.position = rev_z(obj.vertices[v2]) * rate;
							mv2.texcoord = (i+2<uvNum) ? face.uvs[i + 2] : duv;

							node.meshData.vertices.push_back(mv0);
							node.meshData.vertices.push_back(mv1);
							node.meshData.vertices.push_back(mv2);
						}
					}

				}

			}
			else {
				// 炩

				// ɖ@vZ
				for (const auto& v : obj.vertices) {
					MeshVertex mv;
					mv.position = rev_z(v) * rate;
					mv.texcoord = duv;
					node.meshData.vertices.push_back(mv);
				}
				for (const auto& face : obj.faces) {
					int n = face.indices.size();
					if (n >= 3) {
						for (int i = 0; i < n - 2; ++i) {
							int v0 = face.indices[0];
							int v1 = face.indices[i + 1];
							int v2 = face.indices[i + 2];
							node.meshData.indices.push_back(v0);
							node.meshData.indices.push_back(v1);
							node.meshData.indices.push_back(v2);
						}
					}
				}
				node.meshData.computeNormals();

				// ʂ
				for (const auto& face : obj.faces) {
					int matIndex = face.materialIndex;
					if (matIndex < 0 || matIndex >= matNum) {
						continue;
					}
					VertexGroup& gr = node.groups[matIndex];

					bool dbls = f.materials[matIndex].dbls;
					bool addDoubleFace = false;
					if (dbls && addFaceForDouble) {
						addDoubleFace = true;
					}

					int n = face.indices.size();
					int uvNum = face.uvs.size();
					int v0, v1, v2;
					if (n >= 3) {
						// w蒸_ǉ@
						Array<int> adds;
						for (int i = 0; i < n; ++i) {
							v0 = face.indices[i];

							int realIndex = v0;
							if (i < uvNum) {
								auto uv = node.meshData.vertices[v0].texcoord;
								if (uv.x < 0) { // 
									node.meshData.vertices[v0].texcoord = face.uvs[i];
								}
								else if (uv != face.uvs[i]) {
									MeshVertex add(node.meshData.vertices[v0]);
									add.texcoord = face.uvs[i];
									realIndex = node.meshData.vertices.size(); // O
									node.meshData.vertices.push_back(add);
								}
							}
							adds.push_back(realIndex);
						}
						for (int i = 0; i < n - 2; ++i) {
							v0 = adds[0];
							v1 = adds[i+1];
							v2 = adds[i+2];
							gr.indices.push_back(v0);
							gr.indices.push_back(v1);
							gr.indices.push_back(v2);

							if (addDoubleFace) {
								gr.indices.push_back(v2);
								gr.indices.push_back(v1);
								gr.indices.push_back(v0);
							}

						}
					}

				}

			} // ʂ or 炩


			node.meshData.indices.clear();
			for (int i = 0; i < matNum; ++i) {
				node.groups[i].offset = 0;
				if (i >= 1) {
					node.groups[i].offset = node.groups[i - 1].offset
						+ node.groups[i - 1].num();
				}

				for (auto idx : node.groups[i].indices) {
					node.meshData.indices.push_back(idx);
				}
			}
			if (obj.shading == 0) {
				node.meshData.computeNormals();
			}

			this->nodes.push_back(node);

			DynamicMesh dm(node.meshData);
			this->meshes.push_back(dm);
			
			if (dm.isEmpty) {
				//Println(L"̃_Ci~bNbV, ",i);
			}

		} // IuWFNg

		this->computeBaseCoordinates();

		return ret;
	}

	/// <summary>Object̏ԂvZglobal*֔fB
	/// scale͖ 1 1 1 Ƃ݂Ȃ</summary>
	bool computeBaseCoordinates() {
		int num = this->nodes.size();
		for (int i = 0; i < num; ++i) {
			ObjectNode& node = this->nodes[i];
			ObjectNode* p = &this->modelNode;
			int parent = node.parentIndex;
			if (parent >= 0 && parent < num) {
				p = &this->nodes[parent];
			}

			node.globalMov = p->globalMov
				+ p->globalRot * node.translation;
			node.globalRot = p->globalRot * node.rot;
		}
		return true;
	}


	/// <summary>̃CX^XێĂObjectNodeɑΉ\
	/// MotionNodezVɈB</summary>
	MotionInstance CreateMotionInstance() {
		MotionInstance ret;
		ret.modelNode = this->modelNode;
		int num = this->nodes.size();
		for (int i = 0; i < num; ++i) {
			const auto n = this->nodes[i];
			MotionNode node = n;
			ret.nodes.push_back(node);
		}
		return ret;
	}


	/// <summary>̃NXێĂPmdFile֕ϊ</summary>
	/// <param name="f">o͐</param>
	bool ConvertToPmd(PmdFile& f) {
		f.clear();

		Array<int> offsets;

		for (int i = 0; i < this->nodes.size(); ++i) {
			const ObjectNode& node = this->nodes[i];

			// ItZbg͂܂ł̒_
			offsets.push_back(f.vertices.size());

			// _
			int n = node.meshData.vertices.size();
			for (int j = 0; j < n; ++j) {
				const auto& mv = node.meshData.vertices[j];

				PmdVertex v;
				v.position = mv.position;
				v.normal = mv.normal;
				v.texcoord = mv.texcoord;
				v.bones[0] = i;
				v.bones[1] = i;
				v.weight = 100;
				v.edgeFlag = 0; // 0:ʏA1:GbW
				f.vertices.push_back(v);
			}

			// 1IuWFNg1{[
			PmdBone b;
			b.name = node.name;
			b.nameE = Format(L"bone", Pad(f.bones.size(), {5,L'0'}));
			b.parent = node.parentIndex;
			b.child = node.childIndex;
			b.position = node.globalMov;
			b.type = PMDBONE_ROTMOV;
			f.bones.push_back(b);
		}

		for (int i = 0; i < this->mats.size(); ++i) {
			for (int j = 0; j < this->nodes.size(); ++j) {
				if (i < this->nodes[j].groups.size()) {
					const auto& group = this->nodes[j].groups[i];
					for (int k = 0; k < group.indices.size(); ++k) {
						f.faces.push_back(group.indices[k] + offsets[j]);
					}
				}
			}
		}

		for (int i = 0; i < this->mats.size(); ++i) {
			Material& mat = this->mats[i];

			PmdMaterial m;
			m.diffuse = mat.diffuse;
			m.power = mat.shininess;
			m.specular = mat.specular;
			m.ambient = mat.ambient;
			m.texName = mat.diffuseTextureName;

			m.matVertexNum = 0;
			for (int j = 0; j < this->nodes.size(); ++j) {
				const auto& node = this->nodes[j];
				if (i < node.groups.size()) {
					m.matVertexNum += node.groups[i].num();
				}
			}

			f.materials.push_back(m);
		}

		// wb_
		f.name = FileSystem::BaseName(this->path);
		f.nameE = L"Model name";
		f.comment = FileSystem::FileName(this->path) + L" ϊ";
		f.commentE = L"Comment.";

		// {[\g
		int adjustBoneNum = f.bones.size() - 1;
		if (adjustBoneNum < 0) {
			adjustBoneNum = 0;
		}
		int numInFrame = (adjustBoneNum <= 1200) ? 5 : 254;
		int frameNum = (adjustBoneNum + numInFrame - 1) / numInFrame;
		if (frameNum > 255) {
			frameNum = 255; // ő255
		}
		int index = 1; // 擪܂܂Ȃ
		for (int i = 0; i < frameNum; ++i) {
			PmdFrame frame;
			String s = Format(L"bones", Pad(i+1, { 3, L'0' }));
			frame.name = s;
			frame.nameE = s;
			for (int j = 0; j < numInFrame; ++j) {
				if (index >= f.bones.size()) {
					break;
				}
				frame.bones.push_back(index);
				++index;
			}
			f.frames.push_back(frame);
		}

		return true;
	}
};
